/*
 * arch/arm/kernel/hw-ebsa286.c
 *
 * EBSA285 hardware specific functions
 *
 * Copyright (C) 1998 Russell King
 * Copyright (C) 1998, 1999 Phil Blundell
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/pci.h>
#include <linux/ptrace.h>
#include <linux/interrupt.h>
#include <linux/smp.h>
#include <linux/mm.h>
#include <linux/init.h>

#include <asm/dec21285.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/leds.h>
#include <asm/system.h>

#define ETHER10_IO_BASE		0x301
#define DEC21143_IO_BASE	0x401
#define DEC21143_MEM_BASE	0x00800000

extern int setup_arm_irq(int, struct irqaction *);
extern void pci_set_cmd(struct pci_dev *dev, unsigned short clear, unsigned short set);
extern void pci_set_base_addr(struct pci_dev *dev, int idx, unsigned int addr);
extern void pci_set_irq_line(struct pci_dev *dev, unsigned int irq);

static int irqmap_ebsa[] __initdata = { IRQ_IN1, IRQ_IN0, IRQ_PCI, IRQ_IN3 };

__initfunc(static int ebsa_irqval(struct pci_dev *dev))
{
	unsigned char pin;
	
	pcibios_read_config_byte(dev->bus->number,
				 dev->devfn,
				 PCI_INTERRUPT_PIN,
				 &pin);
	
	return irqmap_ebsa[(PCI_SLOT(dev->devfn) + pin) & 3];
}

#ifdef CONFIG_CATS
static int irqmap_cats[] __initdata = { IRQ_PCI, IRQ_IN0, IRQ_IN1, IRQ_IN3 };

__initfunc(static int cats_irqval(struct pci_dev *dev))
{
	if (dev->irq >= 128)
		return 16 + (dev->irq & 0x1f);

	switch (dev->irq) {
	case 1:
	case 2:
	case 3:
	case 4:
		return irqmap_cats[dev->irq - 1];
	case 0:
		return 0;
	}

	printk("PCI: device %02x:%02x has unknown irq line %x\n",
	       dev->bus->number, dev->devfn, dev->irq);
	return 0;
}
#endif

__initfunc(void pcibios_fixup_ebsa285(struct pci_dev *dev))
{
	/* Latency timer of 32 */
	pci_write_config_byte(dev, PCI_LATENCY_TIMER, 32);
	/* 32-byte cache line size */
	pci_write_config_byte(dev, PCI_CACHE_LINE_SIZE, 8);
	/* Set SysErr enable, Parity enable */
	pci_set_cmd(dev, PCI_COMMAND_FAST_BACK, PCI_COMMAND_SERR | PCI_COMMAND_PARITY);

	/* sort out the irq mapping for this device */
	switch (machine_type) {
	case MACH_TYPE_EBSA285:
		dev->irq = ebsa_irqval(dev);
		/* Turn on bus mastering - boot loader doesn't
		 * - perhaps it should! - dag
		 */
		pci_set_cmd(dev, 0, PCI_COMMAND_MASTER);
		break;

#ifdef CONFIG_CATS
	case MACH_TYPE_CATS:
		dev->irq = cats_irqval(dev);
		/* Turn on bus mastering - boot loader doesn't
		 * - perhaps it should! - dag
		 */
		pci_set_cmd(dev, 0, PCI_COMMAND_MASTER);
		break;
#endif
#ifdef CONFIG_ARCH_NETWINDER
	case MACH_TYPE_NETWINDER:
		/* disable ROM */
		pci_write_config_dword(dev, PCI_ROM_ADDRESS, 0);

#define DEV(v,d) ((v)<<16|(d))
		switch (DEV(dev->vendor, dev->device)) {
		/* Ether 100 */
		case DEV(PCI_VENDOR_ID_DEC, PCI_DEVICE_ID_DEC_21142):
			pci_set_base_addr(dev, 0, DEC21143_IO_BASE);
			pci_set_base_addr(dev, 1, DEC21143_MEM_BASE);
			pci_set_cmd(dev, 0, PCI_COMMAND_MASTER | PCI_COMMAND_MEMORY | PCI_COMMAND_IO);
			/* Put the chip to sleep in case the driver isn't loaded */
			pci_write_config_dword(dev, 0x40, 0x80000000);
			dev->irq = IRQ_NETWINDER_ETHER100;
			break;

		/* Ether 10 */
		case DEV(PCI_VENDOR_ID_WINBOND2,0x5a5a):
			pci_set_base_addr(dev, 0, ETHER10_IO_BASE);
			pci_set_cmd(dev, PCI_COMMAND_MASTER | PCI_COMMAND_MEMORY, PCI_COMMAND_IO);
			dev->irq = IRQ_NETWINDER_ETHER10;
			break;

		/* ISA bridge */
		case DEV(PCI_VENDOR_ID_WINBOND,PCI_DEVICE_ID_WINBOND_83C553):
			pci_set_base_addr(dev, 0, 0);
			pci_set_cmd(dev, PCI_COMMAND_MASTER | PCI_COMMAND_MEMORY, PCI_COMMAND_IO);
			/*
			 * Enable all memory requests from ISA to be channeled to PCI
			 */
			pci_write_config_byte(dev, 0x48, 255);
			/*
			 * Disable ping-pong (as per errata)
			 */
			pci_write_config_byte(dev, 0x42, 0);
			/*
			 * Enable PCI packet retry
			 */
			pci_write_config_byte(dev, 0x40, 0x22);
			/*
			 * Do not use PCI CPU park enable, park on
			 * last master, disable GAT bit
			 */
			pci_write_config_byte(dev, 0x83, 0x02);
			/*
			 * Default rotating priorities
			 */
			pci_write_config_byte(dev, 0x80, 0xe0);
			/*
			 * Rotate bank 4
			 */
			pci_write_config_byte(dev, 0x81, 0x01);
			break;

		/* IDE */
		case DEV(PCI_VENDOR_ID_WINBOND,PCI_DEVICE_ID_WINBOND_82C105):
			pci_set_base_addr(dev, 0, 0x1f1);
			pci_set_base_addr(dev, 1, 0x3f5);
			pci_set_base_addr(dev, 2, 0x171);
			pci_set_base_addr(dev, 3, 0x375);
			pci_set_base_addr(dev, 4, 0xe801);
			pci_set_cmd(dev, PCI_COMMAND_MEMORY, PCI_COMMAND_MASTER | PCI_COMMAND_IO);
			dev->irq = IRQ_NETWINDER_IDE;
			break;

		/* VGA */
		case DEV(PCI_VENDOR_ID_INTERG,0x2000):
			pci_set_base_addr(dev, 0, CYBER2000_MEM_BASE);
			pci_set_cmd(dev, PCI_COMMAND_MASTER, PCI_COMMAND_IO | PCI_COMMAND_MEMORY);
			dev->irq = IRQ_NETWINDER_VGA;
			break;
		}
#endif
	}
}

static void
irq_pci_err(int irq, void *dev_id, struct pt_regs *regs)
{
	static unsigned long next_warn;
	unsigned long cmd = *CSR_PCICMD & 0x0000ffff;
	unsigned long ctrl = (*CSR_SA110_CNTL) & 0xffffde07;
	unsigned long irqstatus = *CSR_IRQ_RAWSTATUS;
	struct pci_dev *dev;
	int warn = time_after_eq(jiffies, next_warn);

	if (warn) {
		next_warn = jiffies + 3 * HZ / 100;
		printk(KERN_DEBUG "PCI: ");
	}

	if (irqstatus & (1 << 31)) {
		if (warn)
			printk("parity error ");
		cmd |= 1 << 31;
	}

	if (irqstatus & (1 << 30)) {
		if (warn)
			printk("target abort ");
		cmd |= 1 << 28;
	}

	if (irqstatus & (1 << 29)) {
		if (warn)
			printk("master abort ");
		cmd |= 1 << 29;
	}

	if (irqstatus & (1 << 28)) {
		if (warn)
			printk("data parity error ");
		cmd |= 1 << 24;
	}

	if (irqstatus & (1 << 27)) {
		if (warn)
			printk("discard timer expired ");
		ctrl |= SA110_CNTL_DISCARDTIMER;
	}

	if (irqstatus & (1 << 23)) {
		if (warn)
			printk("system error ");
		ctrl |= SA110_CNTL_RXSERR;
	}

	if (warn)
		printk("pc=%08lX\n", instruction_pointer(regs));

#if 0
	for (dev = pci_devices; dev; dev = dev->next) {
		unsigned short status;

		pci_read_config_word(dev, PCI_STATUS, &status);
		if (status & 0xf900) {
			printk("PCI: [%04X:%04X] status = %X\n",
				dev->vendor, dev->device, status);

			pci_write_config_word(dev, PCI_STATUS, status & 0xf900);
		}
	}
#endif

	*CSR_PCICMD = cmd;
	*CSR_SA110_CNTL = ctrl;
}

static struct irqaction irq_pci_error = {
	irq_pci_err, SA_INTERRUPT, 0, "PCI error", NULL, NULL
};

__initfunc(void pcibios_init_ebsa285(void))
{
	unsigned int mem_size = (int)high_memory - PAGE_OFFSET;
	unsigned long cr;

	cr = *CSR_SA110_CNTL;
	if (cr & (1<<31)) {
		/* We are the PCI configuration master.  Map our memory at 
		   a known address in PCI space, just in case the firmware 
		   had other ideas (or indeed we are running standalone with
		   no supporting firmware at all).  Using a nonzero base for 
		   the SDRAM is a bit un-traditional; allegedly this is
		   necessary because some S3 video cards always decode the
		   ISA memory region at a0000.  */
		*CSR_CSRBASEMASK = 0;
		*CSR_CSRBASEOFFSET = 0;
		*CSR_SDRAMBASEOFFSET = 0;
		*CSR_PCICSRBASE = 0;
		*CSR_PCICSRIOBASE = 0;
		*CSR_PCIROMBASE = 0;
		*CSR_PCICMD = PCI_COMMAND_IO | PCI_COMMAND_MEMORY 
			| PCI_COMMAND_MASTER | PCI_COMMAND_INVALIDATE 
			| PCI_COMMAND_PARITY;
	}
	*CSR_ROMBASEMASK = 0x80000000;
	*CSR_SDRAMBASEMASK = (mem_size - 1) & 0x0ffc0000;
	*CSR_PCISDRAMBASE = virt_to_bus((void *)PAGE_OFFSET);
	*CSR_PCICACHELINESIZE = 0x00002008;
	*CSR_PCIADDR_EXTN = 0;
	*CSR_SA110_CNTL = cr | 1;

	setup_arm_irq(IRQ_PCI_ERR, &irq_pci_error);
}

/*
 * Netwinder stuff
 */
#ifdef CONFIG_ARCH_NETWINDER
/*
 * This is a lock for accessing ports 0x338 and 0x33a
 */
spinlock_t __netwinder_data gpio_lock = SPIN_LOCK_UNLOCKED;

static unsigned int __netwinder_data current_gpio_op = 0;
static unsigned int __netwinder_data current_gpio_io = 0;
static unsigned int __netwinder_data current_cpld = 0;

void __netwinder_text gpio_modify_op(int mask, int set)
{
	unsigned int new_gpio, changed;
unsigned long flags; save_flags_cli(flags);
if ((flags & 128) == 0) {printk("gpio_modify_op called with IRQs enabled from %p\n", __builtin_return_address(0)); __backtrace(); }

	new_gpio = (current_gpio_op & ~mask) | set;
	changed = new_gpio ^ current_gpio_op;
	current_gpio_op = new_gpio;

	if (changed & 0xff)
		outb(new_gpio, 0x338);
	if (changed & 0xff00)
		outb(new_gpio >> 8, 0x33a);
restore_flags(flags);
}

static inline void __gpio_modify_io(int mask, int in)
{
	unsigned int new_gpio, changed;
	int port;

	new_gpio = (current_gpio_io & ~mask) | in;
	changed = new_gpio ^ current_gpio_io;
	current_gpio_io = new_gpio;

	changed >>= 1;
	new_gpio >>= 1;

	outb(7, 0x370);
	outb(7, 0x371);

	for (port = 0xe1; changed && port < 0xe8; changed >>= 1) {
		outb(port, 0x370);
		outb(new_gpio & 1, 0x371);

		port += 1;
		new_gpio >>= 1;
	}

	outb(7, 0x370);
	outb(8, 0x371);

	for (port = 0xe8; changed && port < 0xec; changed >>= 1) {
		outb(port, 0x370);
		outb(new_gpio & 1, 0x371);

		port += 1;
		new_gpio >>= 1;
	}
}

void __netwinder_text gpio_modify_io(int mask, int in)
{
	/* Open up the SuperIO chip */
	outb(0x87, 0x370);
	outb(0x87, 0x370);

	__gpio_modify_io(mask, in);

	/* Close up the EFER gate */
	outb(0xaa, 0x370);
}

int __netwinder_text gpio_read(void)
{
	return inb(0x338) | inb(0x33a) << 8;
}

void __netwinder_text cpld_modify(int mask, int set)
{
	int msk;

	current_cpld = (current_cpld & ~mask) | set;

	gpio_modify_io(GPIO_DATA, 0);
	gpio_modify_op(GPIO_IOLOAD, 0);

	for (msk = 8; msk; msk >>= 1) {
		int bit = current_cpld & msk;

		gpio_modify_op(GPIO_DATA | GPIO_IOCLK, bit ? GPIO_DATA : 0);
		gpio_modify_op(GPIO_IOCLK, GPIO_IOCLK);
	}

	gpio_modify_op(GPIO_IOCLK|GPIO_DATA, 0);
	gpio_modify_op(GPIO_IOLOAD|GPIO_DSCLK, GPIO_IOLOAD|GPIO_DSCLK);
	gpio_modify_op(GPIO_IOLOAD, 0);
}

__initfunc(static void hw_init_cpld(void))
{
	unsigned long flags;

	spin_lock_irqsave(&gpio_lock, flags);
	cpld_modify(-1, CPLD_UNMUTE | 4);
	spin_unlock_irqrestore(&gpio_lock, flags);
}

__initfunc(static void hw_init_wb553(void))
{
	unsigned long flags;
#define WRITE_REG(r,v) do { outb((r), 0x370); outb((v), 0x371); } while (0)

	/* Open up the SuperIO chip */
	outb(0x87, 0x370);
	outb(0x87, 0x370);

	/* Set the serial interrupt numbers */
	WRITE_REG(7, 5);	/* keyboard subsection */
	WRITE_REG(0x72, 5);	/* interrupt number for mouse - IRQ5 */

	WRITE_REG(7, 6);	/* Infrared */
	WRITE_REG(0x70, 6);	/* IRQ 6 */

	WRITE_REG(0x2a, 0xc1);	/* Enable GP12, GP11, GP10 as I/O, CIRRX, IRRXH	*/

	WRITE_REG(0x2b, 0x6b);	/* Enable GP23, GP22, GP21, GP20, GP13 as I/O	*/

	WRITE_REG(0x2c, 0x55);	/* Enable GP17, GP16, GP15, GP14 as I/O		*/

	current_gpio_io = -1;
	__gpio_modify_io(-1, GPIO_DONE | GPIO_WDTIMER);

	WRITE_REG(7, 7);	/* Aux function group 1 (dev 7)			*/
	WRITE_REG(0x60, 0x03);	/* Group controlled by IO port 0x338		*/
	WRITE_REG(0x61, 0x38);	/* GP 11, 12, 13, 14, 15, 16			*/

	WRITE_REG(0x70, 0x0a);	/* IRQ10 for GP10 (Orange button)		*/

	WRITE_REG(0xe0, 0x19);	/* GP10 control reg set for debounce & input	*/
	WRITE_REG(0x30, 0x01);	/* Turn on section 7 (aux function group 1)	*/

	WRITE_REG(7, 8);	/* Aux function group 2 (dev 8)			*/
	WRITE_REG(0x60, 0x03);	/* Group controlled by IO port 0x33a		*/
	WRITE_REG(0x61, 0x3a);

	/* Clear watchdog timer regs */
	WRITE_REG(0xf2, 0x00);	/* Watchdog timeout value (disabled)		*/
	WRITE_REG(0xf3, 0x00);	/* Watchdog reg (reset to default)		*/
	WRITE_REG(0xf4, 0x00);	/* Reset if in timed out state (bit 0)		*/

	/* T.B.D. set IRDA inputs (touch reg 2A, EC, ED) */
	WRITE_REG(0x30, 0x01);	/* Turn on section 8 (aux function group 2)	*/

	spin_lock_irqsave(&gpio_lock, flags);
	gpio_modify_op(-1, GPIO_RED_LED | GPIO_FAN);
	spin_unlock_irqrestore(&gpio_loc, flags);

	/* Close up the EFER gate */
	outb(0xaa, 0x370);
}

static unsigned char rwa_unlock[] __initdata =
{ 0x00, 0x00, 0x6a, 0xb5, 0xda, 0xed, 0xf6, 0xfb, 0x7d, 0xbe, 0xdf, 0x6f, 0x37, 0x1b,
  0x0d, 0x86, 0xc3, 0x61, 0xb0, 0x58, 0x2c, 0x16, 0x8b, 0x45, 0xa2, 0xd1, 0xe8, 0x74,
  0x3a, 0x9d, 0xce, 0xe7, 0x73, 0x39 };

#ifndef DEBUG
#define dprintk if (0) printk
#else
#define dprintk printk
#endif

__initfunc(static void hw_init_rwa010(void))
{
	unsigned char si[9];
	int i, j;

#define WRITE_RWA(r,v) do { outb((r), 0x279); outb((v), 0xa79); } while (0)

	WRITE_RWA(2, 2);
	mdelay(10);

	for (i = 0; i < sizeof(rwa_unlock); i++)
		outb(rwa_unlock[i], 0x279);

	WRITE_RWA(3, 0);
	WRITE_RWA(0, 128);

	outb(1, 0x279);

	mdelay(10);

	dprintk("Identifier: ");
	for (i = 0; i < 9; i++) {
		si[i] = 0;
		for (j = 0; j < 8; j++) {
			int bit;
			mdelay(1);
			inb(0x203);
			mdelay(1);
			bit = inb(0x203);
			dprintk("%02X ", bit);
			si[i] |= bit << j;
		}
		mdelay(10);
		dprintk("%02X ", si[i]);
	}
	dprintk("\n");

	WRITE_RWA(6, 2);	// Assign a card no = 2

	dprintk("Card no = %d\n", inb(0x203));

	WRITE_RWA(7, 3);
	WRITE_RWA(0x30, 0);

	WRITE_RWA(7, 4);
	WRITE_RWA(0x30, 0);

	WRITE_RWA(7, 2);
	WRITE_RWA(0x30, 0);

	WRITE_RWA(7, 5);

	dprintk("Slider base: ");
	WRITE_RWA(0x61, 1);
	i = inb(0x203);

	WRITE_RWA(0x60, 2);
	dprintk("%02X%02X (201)\n", inb(0x203), i);

	WRITE_RWA(0x30, 1);


	WRITE_RWA(7, 0);

	dprintk("WaveArtist base: ");
	WRITE_RWA(0x61, 0x50);
	i = inb(0x203);

	WRITE_RWA(0x60, 0x02);
	dprintk("%02X%02X (250),", inb(0x203), i);

	WRITE_RWA(0x70, 3);
	dprintk(" irq: %d (3),", inb(0x203));

	WRITE_RWA(0x74, 7);
	dprintk(" dma: %d (7)\n", inb(0x203));

	WRITE_RWA(0x30, 1);

	WRITE_RWA(7, 1);

	dprintk("SoundBlaster base: ");
	WRITE_RWA(0x61, 0x20);
	i = inb(0x203);

	WRITE_RWA(0x60, 0x02);
	dprintk("%02X%02X (220),", inb(0x203), i);

	dprintk(" irq: ");
	WRITE_RWA(0x70, 3);
	dprintk("%d (3),", inb(0x203));

	dprintk(" 8-bit DMA: ");
	WRITE_RWA(0x74, 1);
	dprintk("%d (1)\n", inb(0x203));

	dprintk("AdLib base: ");
	WRITE_RWA(0x63, 0x88);
	i = inb(0x203);

	WRITE_RWA(0x62, 0x03);
	dprintk("%02X%02X (388)\n", inb(0x203), i);

	WRITE_RWA(0x30, 1);

	outb(1, 0x226);
	udelay(3);
	outb(0, 0x226);

	for (i = 0; i < 5; i++) {
		if (inb(0x22e) & 0x80)
			break;
		mdelay(1);
	}
	if (i == 5)
		printk("SoundBlaster: DSP reset failed\n");

	dprintk("SoundBlaster DSP reset: %02X (AA)\n", inb(0x22a));

	for (i = 0; i < 5; i++) {
		if ((inb(0x22c) & 0x80) == 0)
			break;
		mdelay(1);
	}

	if (i == 5)
		printk("SoundBlaster: DSP not ready\n");
	else {
		outb(0xe1, 0x22c);

		dprintk("SoundBlaster DSP id: ");
		i = inb(0x22a);
		udelay(1);
		i |= inb(0x22a) << 8;
		dprintk("%04X\n", i);

		for (i = 0; i < 5; i++) {
			if ((inb(0x22c) & 0x80) == 0)
				break;
			mdelay(1);
		}

		if (i == 5)
			printk("SoundBlaster: could not turn speaker off\n");

		outb(0xd3, 0x22c);
	}

	/* turn on opl3 */
	outb(5,0x38a);
	outb(1,0x38b);
}

__initfunc(void hw_init(void))
{
	unsigned long flags;

	hw_init_wb553();
	hw_init_cpld();
	hw_init_rwa010();

#ifdef CONFIG_LEDS /* Clear both LEDs and start the LED driver up */
	spin_lock_irqsave(&gpio_lock, flags);
	gpio_modify_op(GPIO_RED_LED|GPIO_GREEN_LED, 0);
	spin_unlock_irqrestore(&gpio_lock, flags);
	leds_event(led_start);
#else	/* Set the green LED on and the red LED off */
	spin_lock_irqsave(&gpio_lock, flags);
	gpio_modify_op(GPIO_RED_LED|GPIO_GREEN_LED, GPIO_GREEN_LED);
	spin_unlock_irqrestore(&gpio_lock, flags);
#endif /* CONFIG_LEDS */
}

EXPORT_SYMBOL(gpio_lock);
EXPORT_SYMBOL(gpio_modify_op);
EXPORT_SYMBOL(gpio_modify_io);
EXPORT_SYMBOL(cpld_modify);

#endif
